#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "aes.h"

#define COMPOSE(a, b, c, d) (((d)<<24) | ((c)<<16) | ((b)<<8) | (a))
#define DECOMPOSE(x, a, b, c, d) \
	a = (x) & 0xff;\
	b = ((x) >> 8) & 0xff;\
	c = ((x) >> 16) & 0xff;\
	d = ((x) >> 24) & 0xff;

#ifdef _BIG_ENDIAN
#define LE32(x) ((((x)&0xff000000)>>24) | (((x)&0xff0000)>>8) | \
				(((x)&0xff00)<<8) | (((x)&0xff)<<24))
#define NE32(x) ((((x)&0xff000000)>>24) | (((x)&0xff0000)>>8) | \
				(((x)&0xff00)<<8) | (((x)&0xff)<<24))
#else
#define LE32(x) (x)
#define NE32(x) (x)
#endif

#include "aes_tab.h"

static void sub_shift_add(uint32_t v[4], const uint32_t w[4]) {
	unsigned int v00, v01, v02, v03;
	unsigned int v10, v11, v12, v13;
	unsigned int v20, v21, v22, v23;
	unsigned int v30, v31, v32, v33;

	DECOMPOSE(v[0], v00, v01, v02, v03);
	DECOMPOSE(v[1], v10, v11, v12, v13);
	DECOMPOSE(v[2], v20, v21, v22, v23);
	DECOMPOSE(v[3], v30, v31, v32, v33);
	v[0] = w[0] ^ COMPOSE(s_box[v00], s_box[v11], s_box[v22], s_box[v33]);
	v[1] = w[1] ^ COMPOSE(s_box[v10], s_box[v21], s_box[v32], s_box[v03]);
	v[2] = w[2] ^ COMPOSE(s_box[v20], s_box[v31], s_box[v02], s_box[v13]);
	v[3] = w[3] ^ COMPOSE(s_box[v30], s_box[v01], s_box[v12], s_box[v23]);
}

static void sub_shift_mix_add(uint32_t v[4], const uint32_t w[4]) {
	unsigned int v00, v01, v02, v03;
	unsigned int v10, v11, v12, v13;
	unsigned int v20, v21, v22, v23;
	unsigned int v30, v31, v32, v33;

	DECOMPOSE(v[0], v00, v01, v02, v03);
	DECOMPOSE(v[1], v10, v11, v12, v13);
	DECOMPOSE(v[2], v20, v21, v22, v23);
	DECOMPOSE(v[3], v30, v31, v32, v33);
	v[0] = w[0] ^ ssm[0][v00] ^ ssm[1][v11] ^ ssm[2][v22] ^ ssm[3][v33];
	v[1] = w[1] ^ ssm[0][v10] ^ ssm[1][v21] ^ ssm[2][v32] ^ ssm[3][v03];
	v[2] = w[2] ^ ssm[0][v20] ^ ssm[1][v31] ^ ssm[2][v02] ^ ssm[3][v13];
	v[3] = w[3] ^ ssm[0][v30] ^ ssm[1][v01] ^ ssm[2][v12] ^ ssm[3][v23];
}

static void isub_shift_add(uint32_t v[4], const uint32_t w[4]) {
	unsigned int v00, v01, v02, v03;
	unsigned int v10, v11, v12, v13;
	unsigned int v20, v21, v22, v23;
	unsigned int v30, v31, v32, v33;

	DECOMPOSE(v[0], v00, v01, v02, v03);
	DECOMPOSE(v[1], v10, v11, v12, v13);
	DECOMPOSE(v[2], v20, v21, v22, v23);
	DECOMPOSE(v[3], v30, v31, v32, v33);
	v[0] = w[0] ^ COMPOSE(is_box[v00], is_box[v31], is_box[v22], is_box[v13]);
	v[1] = w[1] ^ COMPOSE(is_box[v10], is_box[v01], is_box[v32], is_box[v23]);
	v[2] = w[2] ^ COMPOSE(is_box[v20], is_box[v11], is_box[v02], is_box[v33]);
	v[3] = w[3] ^ COMPOSE(is_box[v30], is_box[v21], is_box[v12], is_box[v03]);
}

static void isub_shift_mix_add(uint32_t v[4], const uint32_t w[4]) {
	unsigned int v00, v01, v02, v03;
	unsigned int v10, v11, v12, v13;
	unsigned int v20, v21, v22, v23;
	unsigned int v30, v31, v32, v33;

	DECOMPOSE(v[0], v00, v01, v02, v03);
	DECOMPOSE(v[1], v10, v11, v12, v13);
	DECOMPOSE(v[2], v20, v21, v22, v23);
	DECOMPOSE(v[3], v30, v31, v32, v33);
	v[0] = w[0] ^ issm[0][v00] ^ issm[1][v31] ^ issm[2][v22] ^ issm[3][v13];
	v[1] = w[1] ^ issm[0][v10] ^ issm[1][v01] ^ issm[2][v32] ^ issm[3][v23];
	v[2] = w[2] ^ issm[0][v20] ^ issm[1][v11] ^ issm[2][v02] ^ issm[3][v33];
	v[3] = w[3] ^ issm[0][v30] ^ issm[1][v21] ^ issm[2][v12] ^ issm[3][v03];
}

static void imix_col(uint32_t v[4]) {
	unsigned int v00, v01, v02, v03;
	unsigned int v10, v11, v12, v13;
	unsigned int v20, v21, v22, v23;
	unsigned int v30, v31, v32, v33;

	DECOMPOSE(v[0], v00, v01, v02, v03);
	DECOMPOSE(v[1], v10, v11, v12, v13);
	DECOMPOSE(v[2], v20, v21, v22, v23);
	DECOMPOSE(v[3], v30, v31, v32, v33);
	v[0] = imix[0][v00] ^ imix[1][v01] ^ imix[2][v02] ^ imix[3][v03];
	v[1] = imix[0][v10] ^ imix[1][v11] ^ imix[2][v12] ^ imix[3][v13];
	v[2] = imix[0][v20] ^ imix[1][v21] ^ imix[2][v22] ^ imix[3][v23];
	v[3] = imix[0][v30] ^ imix[1][v31] ^ imix[2][v32] ^ imix[3][v33];
}

/* Expand a 4*Nk-byte key into w, for use with aes_encrypt_block() */
void aes_expand_key(uint32_t w[Nrk], const void *key) {
	unsigned int t0, t1, t2, t3;
	uint32_t t;
	int i;

	memcpy(w, key, 4*Nk);
	for(i=0; i<Nk; i++) w[i] = LE32(w[i]);
	for(i=Nk; i<Nrk; i++) {
		t = w[i-1];
		if((i % Nk) == 0) {
			DECOMPOSE(t, t0, t1, t2, t3);
			t = COMPOSE(
				s_box[t1] ^ r_con[i/Nk], s_box[t2], s_box[t3], s_box[t0]);
		}
#if Nk == 8
		else if((i % Nk) == 4) {
			DECOMPOSE(t, t0, t1, t2, t3);
			t = COMPOSE(s_box[t0], s_box[t1], s_box[t2], s_box[t3]);
		}
#endif
		w[i] = w[i-Nk] ^ t;
	}
}

/* Convert an expanded encryption key in w to an expanded decryption key in
 * dw. */
void aes_expand_dkey(uint32_t dw[Nrk], const uint32_t w[Nrk]) {
	int i;

	memcpy(dw, w, 4*Nrk);
	for(i=4; i<=(Nr-1)*4; i+=4) {
		imix_col(dw+i);
	}
}

/* Encrypt the 16-byte block in p and put the result in d. w is the expanded
 * encryption key, created with aes_expand_key() */
void aes_encrypt_block(void *d, const void *p, const uint32_t w[Nrk]) {
	uint32_t v[4];
	int i;

	memcpy(v, p, 16);
	for(i=0; i<4; i++) v[i] = w[i] ^ LE32(v[i]);
	for(i=1; i<Nr; i++) {
		w += 4;
		sub_shift_mix_add(v, w);
	}
	sub_shift_add(v, w+4);
	for(i=0; i<4; i++) v[i] = NE32(v[i]);
	memcpy(d, v, 16);
}

/* Decrypt the 16-byte block in p and put the result in d. w is the expanded
 * decryption key, created as follows:
 *
 *		aes_expand_key(w1, key_bytes);
 *		aes_expand_dkey(w, w1);
 */
void aes_decrypt_block(void *d, const void *p, const uint32_t w[Nrk]) {
	uint32_t v[4];
	int i;

	w += Nrk - 4;
	memcpy(v, p, 16);
	for(i=0; i<4; i++) v[i] = w[i] ^ LE32(v[i]);
	for(i=Nr-1; i>=1; i--) {
		w -= 4;
		isub_shift_mix_add(v, w);
	}
	isub_shift_add(v, w-4);
	for(i=0; i<4; i++) v[i] = NE32(v[i]);
	memcpy(d, v, 16);
}

