#include <stdio.h>
#include <string.h>
#include <stdint.h>

static unsigned int mul2(unsigned int x) {
	if((x = x << 1) >= 0x100) return (x & 0xff) ^ 0x1b; else return x;
}

static unsigned int mul(unsigned int x, unsigned int y) {
	unsigned int t = 0;
	while(x) {
		if(x & 1) t = t ^ y;
		y = mul2(y);
		x = x >> 1;
	}
	return t;
}

void build_tab(
	uint32_t tab[4][0x100],
	const uint8_t m[4][4],
	const uint8_t s[0x100]) {

	unsigned int i, j, n;

	for(i=0; i<0x100; i++) {
		n = s[i];
		for(j=0; j<4; j++) {
			tab[j][i] =
				mul(m[0][j], n) |
				(mul(m[1][j], n) << 8) |
				(mul(m[2][j], n) << 16) |
				(mul(m[3][j], n) << 24);
		}
	}
}

static void dump_tab(uint32_t tab[4][0x100], const char *name) {
	unsigned int i, j;
	printf("static const uint32_t %s[4][0x100] = {\n", name);
	for(i=0; i<4; i++) {
		printf("{\n");
		for(j=0; j<0x100; j+=4) {
			printf("\t0x%08x, 0x%08x, 0x%08x, 0x%08x",
				tab[i][j], tab[i][j+1], tab[i][j+2], tab[i][j+3]);
			if(j != 0x100-4) putchar(',');
			putchar('\n');
		}
		putchar('}');
		if(i != 3) putchar(',');
		putchar('\n');
	}
	printf("};\n\n");
}

static void dump_small(uint8_t *tab, unsigned int len, const char *name) {
	unsigned int i,j;

	printf("static const uint8_t %s[0x%x] = {\n", name, len);
	for(i=0; i<len; i+=8) {
		putchar('\t');
		for(j=0; j<8 && i+j<len; j++) {
			printf("0x%02x", tab[i+j]);
			if(i+j<len-1) printf(", ");
		}
		putchar('\n');
	}
	printf("};\n\n");
}

int main(void) {
	uint32_t tab[4][0x100];
	const uint8_t m[4][4] = {
		{ 0x02, 0x03, 0x01, 0x01 },
		{ 0x01, 0x02, 0x03, 0x01 },
		{ 0x01, 0x01, 0x02, 0x03 },
		{ 0x03, 0x01, 0x01, 0x02 }
	};
	const uint8_t im[4][4] = {
		{ 0x0e, 0x0b, 0x0d, 0x09 },
		{ 0x09, 0x0e, 0x0b, 0x0d },
		{ 0x0d, 0x09, 0x0e, 0x0b },
		{ 0x0b, 0x0d, 0x09, 0x0e }
	};
	uint8_t id[0x100];
	uint8_t exp_tab[0x100], log_tab[0x100], inv_tab[0x100];
	uint8_t rcon_tab[0xff], sbox_tab[0x100], isbox_tab[0x100];
	uint8_t s,t;
	int i,j;

	for(i=0; i<0x100; i++) id[i] = i;

	exp_tab[0] = 1;
	for(i=1; i<0x100; i++) {
		exp_tab[i] = mul(3, exp_tab[i-1]);
	}
	log_tab[0] = 0;
	for(i=1; i<0x100; i++) {
		log_tab[exp_tab[i]] = i;
	}
	inv_tab[0] = 0;
	inv_tab[1] = 1;
	for(i=1; i<0x100; i++) {
		inv_tab[i] = exp_tab[0xff - log_tab[i]];
	}

	for(i=0; i<0x100; i++) {
		t = inv_tab[i];
		s = 0x63 ^ t;
		for(j=0; j<4; j++) {
			t = (t << 1) | (t >> 7);
			s = s ^ t;
		}
		sbox_tab[i] = s;
	}
	dump_small(sbox_tab, 0x100, "s_box");

	for(i=0; i<0x100; i++) {
		isbox_tab[sbox_tab[i]] = i;
	}
	dump_small(isbox_tab, 0x100, "is_box");

	rcon_tab[1] = 1;
	for(i=2; i<0xff; i++) rcon_tab[i] = mul2(rcon_tab[i-1]);
	rcon_tab[0] = mul2(rcon_tab[0xfe]);
	dump_small(rcon_tab, 0xff, "r_con");

	build_tab(tab, m, sbox_tab);
	dump_tab(tab, "ssm");

	build_tab(tab, im, isbox_tab);
	dump_tab(tab, "issm");

	build_tab(tab, im, id);
	dump_tab(tab, "imix");

	return 0;
}

